//------------------------------------------------------------------------------
// Bearbeitungsstand: MIDI_Player_TFT_V1_0_2SoftSPI_PulseTest1 28.August 2021
//Tempo Adjust über Button "Up" / "Down" eingefügt
//optimierte TFT Ausgabe
//Version mit SoftSPI für ext. SD-Card
//CycleTime liegt bei 12ms (ok)
//Test mit Kurbelabfrage über Encoder, aktivieren über Pin 35 LOW, Pin 31=A, Pin 33=B (noch anpassen)
//Progressbar eingefügt (experimentell, noch zu programmieren)
//Start Play extern über "Münzeinwurf" Pin 37 LOW, Abfrage in "read_Button" eingefügt (ok)
//Testaufbau Verdrahtung komplettiert
//Batterie Status eingefügt. Zellenspannung 3,15V=0%, 4,06V=100% entspricht einer Batteriespannung von 9,45V - 12,18V
//Batterie Symbol mit Balkenanzeige eingefügt (ok)
//MIDI-Eingang, wenn keine SD-Card eingelegt (ok)
//Tuner eingefügt noch zu ändern
//TFT TouchScreen funktioniert nicht mit ADC??
//------------------------------------------------------------------------------
#include <Adafruit_GFX.h>
#include <MCUFRIEND_kbv.h>
#include <SdFat.h>
#include <MD_MIDIFile.h>
#include <MD_MIDIHelper.h>

MCUFRIEND_kbv tft;
MD_MIDIFile SMF;
#include <Fonts/FreeSans9pt7b.h>
#include <Fonts/FreeSans12pt7b.h>
#include <Fonts/FreeSerif12pt7b.h>
#include <FreeDefaultFonts.h>


#include <TouchScreen.h>

//------------------------------------------------------------------------------
#include <Arduino.h>
#include <MD_REncoder.h>
MD_REncoder R = MD_REncoder(33, 31);
//------------------------------------------------------------------------------

//--------------------------------------------------------------------------
//BatterieStatus
#include "Ticker.h"                                 //Zeitgesteuerte Funktionen (Intervall) https://github.com/sstaub/Ticker
#include "RunningMedian.h"                          //Mittelwert/Durchschnitt von Messergebnissen https://github.com/RobTillaart/Arduino/tree/master/libraries/RunningMedian
const byte BatPin = A7;                             //Analogeingang über Spannungsteiler 1/2,52

void getBattery();                                  //Funktionsdeklaration
Ticker timer1(getBattery, 1000);                    //Funktion alle X ms

RunningMedian median1 = RunningMedian(10);          //Mittelwert mittels X Durchläufen ermitteln
//----------------------------------------------------------------------------




#define MINPRESSURE 200
#define MAXPRESSURE 1000

#define DEBUG_ON 1

#define DEBUG(x)  Serial.println(x)
#define DEBUGX(x) Serial.print(x)
#define DEBUGb(x) Serial.println(x, BIN)



//------------------------------------------------------------------------------


#if SPI_DRIVER_SELECT == 2  // Must be set in SdFat/SdFatConfig.h
#define SD_FAT_TYPE 0
//Für SD-Card auf TFT-Board
//const uint8_t SD_CS_PIN = 10;
//const uint8_t SOFT_MISO_PIN = 12;
//const uint8_t SOFT_MOSI_PIN = 11;
//const uint8_t SOFT_SCK_PIN  = 13;

//Für ext. SD-Card mit Mega Zither
const uint8_t SD_CS_PIN = 53;
const uint8_t SOFT_MISO_PIN = 50;
const uint8_t SOFT_MOSI_PIN = 51;
const uint8_t SOFT_SCK_PIN  = 52;

//Für ext. SD-Card mit Mega Test
//const uint8_t SD_CS_PIN = 53;
//const uint8_t SOFT_MISO_PIN = 47;
//const uint8_t SOFT_MOSI_PIN = 49;
//const uint8_t SOFT_SCK_PIN  = 51;



bool pressed = 0;

// Playlist handling -----------
const uint8_t FNAME_SIZE = 13;               // file names 8.3 to fit onto TFT display
const char* PLAYLIST_FILE = "PLAYLIST.TXT";  // file of file names
const char* MIDI_EXT = ".mid";               // MIDI file extension
uint16_t  plCount = 0;
char fname[FNAME_SIZE + 1];
static uint16_t  cycleTime;

unsigned long duration;
int pin = 35;//Auswahl mit/ohne Kurbelsensor
int coin = 37;//Münzeinwurf
const uint8_t BEAT_LED1 = 43;
const uint8_t BEAT_LED2 = 45;
uint16_t elapsedTime = 0;

// Enumerated types for the FSM(s)
enum tft_state { LSBegin, LSSelect, LSShowFile };
enum midi_state { MSBegin, MSLoad, MSOpen, MSProcess, MSClose };
enum seq_state { TFTSeq, MIDISeq };

uint16_t  lbl = 0;// BACK,STOP,PLAY,PAUSE,VWD,UP,DOWN

SoftSpiDriver<SOFT_MISO_PIN, SOFT_MOSI_PIN, SOFT_SCK_PIN> softSpi;

#if ENABLE_DEDICATED_SPI
#define SD_CONFIG SdSpiConfig(SD_CS_PIN, DEDICATED_SPI, SD_SCK_MHZ(0), &softSpi)
#else  // ENABLE_DEDICATED_SPI
#define SD_CONFIG SdSpiConfig(SD_CS_PIN, SHARED_SPI, SD_SCK_MHZ(0), &softSpi)
#endif  // ENABLE_DEDICATED_SPI

#if SD_FAT_TYPE == 0
SdFat sd;
File file;
#elif SD_FAT_TYPE == 1
SdFat32 sd;
File32 file;
#elif SD_FAT_TYPE == 2
SdExFat sd;
ExFile file;
#elif SD_FAT_TYPE == 3
SdFs sd;
FsFile file;
#else  // SD_FAT_TYPE
#error Invalid SD_FAT_TYPE
#endif  // SD_FAT_TYPE

//------------------------------------------------------------------------------

static uint8_t plIndex = 0;
static SdFile plFile;  // play list file
//------------------------------------------------------------------------------

// Werte aus Kalibrierung mit TouchScreen_Calibr_native.ino
const int XP = 6, XM = A2, YP = A1, YM = 7; //ID=0x9486
const int TS_LEFT = 955, TS_RT = 174, TS_TOP = 167, TS_BOT = 928;
TouchScreen ts = TouchScreen(XP, YP, XM, YM, 300);

Adafruit_GFX_Button tune_btn, midi_btn, play_btn, stop_btn, back_btn, vwd_btn, pause_btn, up_btn, down_btn, ret_btn, minus_btn, plus_btn, ret2_btn ;

int pixel_x, pixel_y;     //Touch_getXY() updates global vars
bool Touch_getXY(void)
{

  TSPoint p = ts.getPoint();


  pinMode(YP, OUTPUT);      //restore shared pins
  pinMode(XM, OUTPUT);
  digitalWrite(YP, HIGH);   //because TFT control pins
  digitalWrite(XM, HIGH);
  //bool pressed = (p.z > MINPRESSURE && p.z < MAXPRESSURE);
  bool pressed = (p.z > MINPRESSURE && p.z < MAXPRESSURE);

  if (pressed) {
    pixel_x = map(p.y, TS_LEFT, TS_RT, 0, tft.width()); //.kbv makes sense to me
    pixel_y = map(p.x, TS_TOP, TS_BOT, 0, tft.height());

  }

  return pressed;
}
/*  r     g    b */
#define BLACK        0x0000  /*   0,   0,   0 */
#define BLUE         0x001F  /*   0,   0, 255 */
#define RED          0xF800  /* 255,   0,   0 */
#define GREEN        0x07E0  /*   0, 255,   0 */
#define CYAN         0x07FF  /*   0, 255, 255 */
#define MAGENTA      0xF81F  /* 255,   0, 255 */
#define YELLOW       0xFFE0  /* 255, 255,   0 */
#define WHITE        0xFFFF  /* 255, 255, 255 */
#define NAVY         0x000F  /*   0,   0, 128 */
#define DARKGREEN    0x03E0  /*   0, 128,   0 */
#define DARKCYAN     0x03EF  /*   0, 128, 128 */
#define MAROON       0x7800  /* 128,   0,   0 */
#define PURPLE       0x780F  /* 128,   0, 128 */
#define OLIVE        0x7BE0  /* 128, 128,   0 */
#define LIGHTGREY    0xC618  /* 192, 192, 192 */
#define DARKGREY     0x7BEF  /* 128, 128, 128 */
#define ORANGE       0xFD20  /* 255, 165,   0 */
#define GREENYELLOW  0xAFE5  /* 173, 255,  47 */
#define GREY         0x8410

//----------------------------------------------------------------------------

//Tuner Code Start
#include <avr/pgmspace.h>
#include <EEPROM.h>

// from wiring_private.h
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif



extern const uint8_t PROGMEM digital_pin_to_bit_mask_PGM[];
extern const uint8_t PROGMEM digital_pin_to_timer_PGM[];
#define digitalPinToPort(P) ( pgm_read_byte( digital_pin_to_port_PGM + (P) ) )
#define digitalPinToBitMask(P) ( pgm_read_byte( digital_pin_to_bit_mask_PGM + (P) ) )
#define digitalPinToTimer(P) ( pgm_read_byte( digital_pin_to_timer_PGM + (P) ) )


static int const kStatusPin = 37;
static int const kStatusPin2 = 39;
static int const kPitchDownButtonPin = 31;
static int const kPitchUpButtonPin = 33;
static int const kModeButtonPin = 35;

#include <inttypes.h>
#include "Print.h"

static int const CHARACTER_WIDTH = 5;
static int const DISPLAY_WIDTH = 16;
static int const MAX_TICK = CHARACTER_WIDTH * (DISPLAY_WIDTH - 1);
uint16_t lastpitchBarX;
///////////////////////////////////////////////////////////////////////////////
// Utility stuff
///////////////////////////////////////////////////////////////////////////////

// From http://www.arduino.cc/playground/Code/AvailableMemory
// This function will return the number of bytes currently free in RAM.
// written by David A. Mellis
// based on code by Rob Faludi http://www.faludi.com
// (reformatted for clarity)
// Note: I believe this does not work on recent versions of the Arduino environment.
// See the webpage above for alternatives.
int AvailableMemory()
{
  int size = 1024;
  byte* buf;

  while ((buf = (byte *) malloc(--size)) == NULL);
  free(buf);
  return size;
}

const int kDebounceMs = 50;

void Blink(int times = 1)
{

  for (int i = 0; i < times; ++i)
  {
    digitalWrite(kStatusPin, HIGH);
    delay(kDebounceMs);
    digitalWrite(kStatusPin, LOW);
    delay(kDebounceMs);
  }
}

class PushButton
{
  public:
    PushButton(int pin, bool activeLow = true)
      : m_pin(pin)
      , m_debounceStart(0)
      , m_activeLow(activeLow)
    {
      pinMode(m_pin, INPUT_PULLUP);
    }

    void Update()
    {
      m_justPressed = false;
      m_justReleased = false;

      if (millis() < m_debounceStart + kDebounceMs)
      {
        return;
      }

      int buttonValue = digitalRead(m_pin);
      int const ACTIVE = m_activeLow ? 0 : 1;
      int const INACTIVE = 1 - ACTIVE;
      if ((m_lastValue == INACTIVE) && (buttonValue == ACTIVE))
      {
        m_debounceStart = millis();
        m_justPressed = true;
      }
      else if ((m_lastValue == ACTIVE) && (buttonValue == INACTIVE))
      {
        m_debounceStart = millis();
        m_justReleased = true;
      }
      m_lastValue = buttonValue;
    }

    bool IsPressed() {
      return m_lastValue == (m_activeLow ? 0 : 1);
    }
    bool JustPressed() {
      return m_justPressed;
    }
    bool JustReleased() {
      return m_justReleased;
    }

    void WaitForPress()
    {
      for (; !JustPressed(); Update());
      Update();

    }

  private:
    int m_pin;
    int m_lastValue;
    long m_debounceStart;
    bool m_activeLow;
    bool m_justPressed;
    bool m_justReleased;
};

PushButton g_pitchDownButton(kPitchDownButtonPin);
PushButton g_pitchUpButton(kPitchUpButtonPin);
PushButton g_modeButton(kModeButtonPin);

///////////////////////////////////////////////////////////////////////////////
// The Tuner
///////////////////////////////////////////////////////////////////////////////
// Note Freq    Cycles
// A  110     145454.55
// A# 116.54  137290.81
// B  123.47  129585.27
// C  130.81  122312.21
// C# 138.59  115447.35
// D  146.83  108967.79
// D# 155.56  102851.9
// E  164.81  97079.26
// F  174.61  91630.62
// F# 185     86487.79
// G  196     81633.6
// G# 207.65  77051.86
// A  220     72727.27

static int const NUM_NOTES = 13;
static char* g_noteNames[NUM_NOTES] =
{
  "A ",
  "A#",
  "B ",
  "C ",
  "C#",
  "D ",
  "D#",
  "E ",
  "F ",
  "F#",
  "G ",
  "G#",
  "A "
};

int g_noteGlyphIndex[NUM_NOTES] =
{
  0,
  0,
  1,
  2,
  2,
  3,
  3,
  4,
  5,
  5,
  6,
  6,
  0
};

bool g_noteSharpSign[NUM_NOTES] =
{
  false,
  true,
  false,
  false,
  true,
  false,
  true,
  false,
  false,
  true,
  false,
  true,
  false
};

static int const A440_NOTE = 69;
static int const A880_NOTE = A440_NOTE + 12;
float const QUARTERTONE_UP = 1.0293022366f;
float const QUARTERTONE_DOWN = 0.9715319412f;
float const SEMITONE_UP = 1.0594630944f;
float const SEMITONE_DOWN = 0.9438743127f;

typedef int Fixed;
static int const FIXED_SHIFT = 5;
static int const FIXED_ONE = (1 << FIXED_SHIFT);
static int const FRAC_MASK = (1 << FIXED_SHIFT) - 1;
static int const FIXED_SHIFT_HI = 8 - FIXED_SHIFT;
static int const FIXED_MASK_HI = ~((1 << FIXED_SHIFT_HI) - 1);
#define FIXED_INT(x) int(x >> FIXED_SHIFT)
#define FIXED_FRAC(x) int(x & FRAC_MASK)
#define   I2FIXED(x) ((Fixed) ((x) << FIXED_SHIFT))
#define   F2FIXED(x) ((Fixed) ((x) * (1 << FIXED_SHIFT)))
#define   FIXED2I(x) ((x) >> FIXED_SHIFT)
#define   FIXED2F(x) ((x) / float(1 << FIXED_SHIFT))

// Useful to print compile-time integer values, since instanciating a template with this incomplete type will
// cause the compiler to include the value of N in the error message.  (On avr-gcc, this seems to only work
// with free-standing variables, not members.)
template<int N> struct PrintInt;

// Linearly interpolates between two 8-bit signed values.
char InterpolateChar(char a, char b, char tFrac)
{
  int d = b - a;
  return a + ((d * tFrac) >> FIXED_SHIFT);
}

namespace TunerMode
{
enum Type
{
  Tuner = 0,
  Midi,
  Max
};
}

class Tuner
{
  public:
    Tuner(int audioPin)
      : m_audioPin(audioPin)

      , m_midiNote(-1)
      , m_a440(440.0f)
      , m_mode(TunerMode::Tuner)

    {

    }

    ~Tuner()
    {


    }

    static int const PRESCALER = 0b00000111;
    static int const PRESCALER_DIVIDE = (1 << PRESCALER);
    static int const ADC_CLOCKS_PER_ADC_CONVERSION = 13;
    static unsigned long const CPU_CYCLES_PER_SAMPLE = ADC_CLOCKS_PER_ADC_CONVERSION * PRESCALER_DIVIDE;
    static unsigned long const SAMPLES_PER_SECOND = F_CPU / CPU_CYCLES_PER_SAMPLE;

    static int const MIN_FREQUENCY = 60;
    static int const MAX_FREQUENCY = 880;
    static int const MIN_SAMPLES = SAMPLES_PER_SECOND / MAX_FREQUENCY;
    static int const MAX_SAMPLES = SAMPLES_PER_SECOND / MIN_FREQUENCY;
    static int const WINDOW_SIZE = MAX_SAMPLES; //96; // samples
    static int const BUFFER_SIZE = WINDOW_SIZE + MAX_SAMPLES + 1; // for interpolation
    static int const AMPLITUDE_THRESHOLD = 30;
    static int const CORRELLATION_STEP = 2;
    // For pure sine, 8 is better than 10 or 12, which causes octave errors at higher frequencies
    // (>400 for 10, >350 for 12).
    // For guitar, sweet spot is between 1/2 and 1/3
    static int const PRIME_THRESHOLD = (FIXED_ONE * 5) / 12;
    static int const OFFSET_STEP = 1;

    void Start()
    {



      // Enable auto-trigger enable
      sbi(ADCSRA, ADATE);
      // Set auto-trigger to free-running mode
      cbi(ADCSRB, ADTS0);
      cbi(ADCSRB, ADTS1);
      cbi(ADCSRB, ADTS2);

      // Select input channel + set reference to Vcc

      //ADMUX = /*(0 << 6) |*/ (m_audioPin & 0x0f);

      ADMUX = (1 << 6) | (m_audioPin & 0x0f);

      // Disable other ADC channels (try to reduce noise?)
      DIDR0 = (0x3F ^ (1 << m_audioPin));

      // Left-adjust result so we only have to read 8 bits
      sbi(ADMUX, ADLAR); // right-adjust for 8 bits

      // Setup the prescaler; divide by 32
      unsigned char adcsra = ADCSRA;

      adcsra = ((adcsra & 0xF8) | PRESCALER); // mask off / re-set prescaler bits
      ADCSRA = adcsra;

      // Disable the conversion complete interrupt so we can read the flag
      cbi(ADCSRA, ADIE); // Interrupt Enable

      // Start the shebang
      sbi(ADCSRA, ADSC);


      LoadTuning();
      TunePitch();

    }

    void Stop()
    {
      
      // Disable auto-trigger mode
      cbi(ADCSRA, ADATE);

      // Right-adjust result (default)
      cbi(ADMUX, ADLAR);


    }

    unsigned long LoadEepromLong(int address)
    {
      unsigned long result = 0;
      unsigned long t = 0; // to avoid spurious warning :/
      result |= EEPROM.read(address);
      t = EEPROM.read(address + 1);
      result |= (t << 8);
      t = EEPROM.read(address + 2);
      result |= (t << 16);
      t = EEPROM.read(address + 3);
      result |= (t << 24);
      return result;
    }

    void SaveEepromLong(int address, unsigned long i)
    {
      EEPROM.write(address, i & 0xFF);
      EEPROM.write(address + 1, (i >> 8) & 0xFF);
      EEPROM.write(address + 2, (i >> 16) & 0xFF);
      EEPROM.write(address + 3, (i >> 24) & 0xFF);
    }

    void SaveTuning()
    {

      SaveEepromLong(0, *((unsigned long*) &m_a440));
    }

    void LoadTuning()
    {

      *((unsigned long*) &m_a440) = LoadEepromLong(0);
    }

    // The following two functions require proper setup in Start()
    unsigned int ReadInput8BitsUnsigned()
    {
      while ((ADCSRA & _BV(ADIF)) == 0)
      {
      }
      unsigned int result = ADCH;
      sbi(ADCSRA, ADIF);
      return result;
    }

    int ReadInput8BitsSigned()
    {
      return ReadInput8BitsUnsigned() - 128;
    }

    // Given a MIDI note index, returns the corresponding index into the global string array of note names.
    int GetNoteNameIndex(int note)
    {
      return (note + 3) % 12;
    }

    // Given a MIDI note index, get the name of the note as ASCII.
    char* GetNoteName(int note)

    {
      if (note >= 0)
      {
        return g_noteNames[GetNoteNameIndex(note)];
      }
      else
      {
        return "  ";
      }
    }

    // Converts a fundamental frequency in Hz to a MIDI note index.  Slow.
    int GetMidiNoteIndexForFrequency(float frequency)
    {
      if (frequency < 0.0f)
      {
        return -1;
      }

      frequency *= QUARTERTONE_DOWN; // kind of like rounding the note, half a semitone (it's a multiplication because of logarithmic space)

      int note = A440_NOTE;
      float a440 = m_a440;

      float a440_2 = 2.0f * a440;
      while (frequency < a440)
      {
        frequency *= 2.0f;
        note -= 12;
      }
      while (frequency > a440_2)
      {
        frequency *= 0.5f;
        note += 12;
      }
      while (frequency > a440)
      {
        frequency *= SEMITONE_DOWN;
        note += 1;
      }

      return note;
    }

    // Compute the fundamental frequency of a given MIDI note index.  Slow.
    float GetFrequencyForMidiNoteIndex(int note)
    {
      if (note < 0.0f)
      {
        return -1.0f;
      }

      float result = m_a440;

      while (note < A440_NOTE)
      {
        note += 12;
        result *= 0.5f;
      }

      while (note > A880_NOTE)
      {
        note -= 12;
        result *= 2.0f;
      }

      while (note > A440_NOTE)
      {
        --note;
        result *= SEMITONE_UP;
      }

      return result;
    }

    unsigned long GetCorrellationFactorFixed(char* buffer, Fixed fixedOffset)
    {
      unsigned long result = 0;
      int integer = FIXED_INT(fixedOffset);
      int frac = FIXED_FRAC(fixedOffset);
      int correllationStep = CORRELLATION_STEP;

      // If we're in MIDI mode, lower the precision to gain speed.
      if (m_mode == TunerMode::Midi)
      {
        correllationStep <<= 1;
      }

      for (int i = 0; i < WINDOW_SIZE; i += correllationStep)
      {
        // Note this is done with 16-bit math; this is slower, but gives more precision.  In tests, using 8-bit
        // math did not yield sufficient precision.
        int a = buffer[i];
        int b = InterpolateChar(buffer[i + integer], buffer[i + integer + 1], frac);
        result += abs(b - a);
      }
      return result;
    }

    unsigned long GetCorrellationFactorPrime(unsigned long currentCorrellation, int numToDate, unsigned long sumToDate)
    {
      if (numToDate == 0)
      {
        return FIXED_ONE;
      }
      else
      {
        return ((currentCorrellation << FIXED_SHIFT) * numToDate) / sumToDate;
      }
    }

    // Compute the frequency corresponding to a given a fixed-point offset into our sampling buffer (usually where
    // the best/minimal autocorrellation was achieved).
    float GetFrequencyForOffsetFixed(Fixed offset)
    {
      float floatOffset = FIXED2F(offset);
      return F_CPU / (floatOffset * CPU_CYCLES_PER_SAMPLE);
    }

    float DetermineSignalPitch()
    {
      // Sample the signal into our buffer, and track its amplitude.
      m_maxAmplitude = 0;
      for (int i = 0; i < BUFFER_SIZE; ++i)
      {
        m_buffer[i] = ReadInput8BitsSigned();
        if (abs(m_buffer[i]) > m_maxAmplitude)
        {
          m_maxAmplitude = abs(m_buffer[i]);
        }
      }

      float result = 0.0f;

      // If we haven't reached the amplitude threshold, don't try to determine pitch.
      if (m_maxAmplitude < AMPLITUDE_THRESHOLD)
      {
        result = -1.0f;
      }

      // Alright, now try to figure what the ballpark note this is by calculating autocorrellation
      Fixed bestOffset = 0;
      Fixed incrementAtBestOffset = 0;

      if (result >= 0.0f)
      {
        int numCorrellations = 0;
        unsigned long sumToDate = 0;
        bool inThreshold = false;
        unsigned long bestPrime = ~0;
        Fixed offsetStep = OFFSET_STEP;

        // If we're in MIDI mode, double the step; this will reduce precision, but increase speed (and thus reduce
        // latency).
        if (m_mode == TunerMode::Midi)
        {
          offsetStep <<= 1;
        }

        Fixed offsetIncrement = offsetStep;
        for (Fixed offset = I2FIXED(MIN_SAMPLES); FIXED_INT(offset) < MAX_SAMPLES; offset += offsetIncrement)
        {
          unsigned long curCorrellation = GetCorrellationFactorFixed(m_buffer, offset);
          ++numCorrellations;
          sumToDate += curCorrellation;

          // Increment the increment right away, so we save the range appropriately for the refined search
          offsetIncrement += offsetStep;

          unsigned long prime = GetCorrellationFactorPrime(curCorrellation, numCorrellations, sumToDate);
          if (prime < bestPrime)
          {
            bestPrime = prime;
            bestOffset = offset;
            incrementAtBestOffset = offsetIncrement;
          }

          if (prime < PRIME_THRESHOLD)
          {
            inThreshold = true;
          }
          else if (inThreshold) // was in threshold, now exited, have best minimum in threshold
          {
            //found = true;
            break;
          }
        }
      }

      if (FIXED_INT(bestOffset) == MAX_SAMPLES)
      {
        result = -1.0f;
      }

#define SUBDIVIDE 1
#if SUBDIVIDE
      // If we're in tuner mode, try to refine the pitch estimate by interpolating samples, for subsample accuracy.
      if ((result >= 0.0f) && (m_mode == TunerMode::Tuner))
      {
        // Upsample the signal to get a better bearing on the real frequency
        Fixed minSamples = bestOffset - incrementAtBestOffset;
        Fixed maxSamples = bestOffset + incrementAtBestOffset;
        unsigned long bestCorrellation = ~0;
        bestOffset = 0;
        for (Fixed offset = minSamples; offset <= maxSamples; ++offset) // step by one
        {
          unsigned long curCorrellation = GetCorrellationFactorFixed(m_buffer, offset);
          if (curCorrellation < bestCorrellation)
          {
            bestCorrellation = curCorrellation;
            bestOffset = offset;

          }
        }
      }
#endif

      int const T1 = AMPLITUDE_THRESHOLD;
      int const T2 = 120;


      if (m_maxAmplitude > T1)

      {

        if (result >= 0.0f)
        {
          // If we hit the end, assume no periodicity
          result = GetFrequencyForOffsetFixed(bestOffset);
        }
        digitalWrite(kStatusPin, HIGH);
        tft.fillCircle(140, 120, 8, GREEN);

        char  sBuf[10];
        sprintf(sBuf, "%3d", m_maxAmplitude);
        showmsgXY(120, 140, 2, NULL, sBuf, WHITE);

        char  sBuf1[10];
        sprintf(sBuf1, "%3d", T2);
        showmsgXY(310, 140, 2, NULL, sBuf1, WHITE);
      }
      else
      {
        result = -1.0f;
        digitalWrite(kStatusPin, LOW);
        tft.fillCircle(140, 120, 8, BLACK);
        tft.drawCircle(140, 120, 8, RED);
      }

      if (m_maxAmplitude > T2)
      {
        digitalWrite(kStatusPin2, HIGH);
        tft.fillCircle(330, 120, 8, RED);
      }
      else
      {
        digitalWrite(kStatusPin2, LOW);
        tft.fillCircle(330, 120, 8, BLACK);
        tft.drawCircle(330, 120, 8, GREEN);
      }

      return result;
    }

    void RenderWideGlyphForNote(int note)
    {

      if (note >= 0)
      {
        int nameIndex = GetNoteNameIndex(note);
        int glyphIndex = g_noteGlyphIndex[nameIndex];

        showmsgXY(210, 110, 7, NULL, (GetNoteName(note)), WHITE);// gibt Note in Klartext auf dem TFT aus






        if (g_noteSharpSign[nameIndex])
        {

        }
        else
        {

        }
      }
      else
      {

      }

    }

    void PrintCenterTick()
    {
      //hier die Skala und Mittenmarkierung erstellen
      for (int i = 100; i < 380; i = i + 28) {
        tft.drawFastVLine(i, 210, 15, WHITE);
      }

      for (int i = 100; i < 380; i = i + 14) {
        tft.drawFastVLine(i, 210, 7, WHITE);
      }
      tft.fillTriangle(235, 225, 240, 210, 245, 225, RED);
      showmsgXY(110, 230, 2, NULL, "-4", WHITE);
      showmsgXY(345, 230, 2, NULL, "+4", WHITE);
    }

    void TFT_Tick(float f)
    {
      uint16_t tick = uint16_t(f * MAX_TICK);
      uint16_t x = (tick / 5) + 1;
      uint16_t glyph = tick % 5;

      Status4(x, tick, glyph);
    }

    void NoteOn()
    {
      Serial.write(0x90); // note on, channel 0
      Serial.write(char(m_midiNote));
      Serial.write(0x40); // velocity 64


    }

    void NoteOff()
    {
      if (m_midiNote >= 0)
      {
        Serial.write(0x80); // note off, channel 0;
        Serial.write(char(m_midiNote));
        Serial.write(0x7F); // velocity 127


        m_midiNote = -1;
      }
    }

    void TunePitch()
    {

      unsigned long lastPress = millis();
      bool firstTime = true;
      while (millis() - lastPress < 2000)
      {
               
        g_pitchDownButton.Update();
        g_pitchUpButton.Update();

        bool buttonPressed = firstTime;
        if (g_pitchDownButton.JustPressed())
        {
          m_a440 -= 1.0f;
          //m_a440 = 440.0f;

          buttonPressed = true;
        }

        if (g_pitchUpButton.JustPressed())
        {
          m_a440 += 1.0f;

          buttonPressed = true;
        }

        if (m_a440 < MIN_FREQUENCY)
        {
          m_a440 = MIN_FREQUENCY;
        }

        if (m_a440 > MAX_FREQUENCY)
        {
          m_a440 = MAX_FREQUENCY;

        }

        if (buttonPressed)
        {
          lastPress = millis();

          char  sBuf[20];
          int y = (m_a440);
          sprintf(sBuf, "Tune %d Hz", y);
          Serial.print(sBuf);
          showmsgXY(180, 230, 2, NULL, sBuf, WHITE);



        }

        firstTime = false;
      }

      SaveTuning();
    }//Ende void TunePitch()

    void Go()
    {

      Start();

      int i = 0;
      float filteredFrequency = -1.0f;
      int noteRepeatCount = 0;

      while (1)
      {
        

        
        g_pitchDownButton.Update();
        g_pitchUpButton.Update();

        if (g_pitchDownButton.IsPressed() || g_pitchUpButton.IsPressed())
        {
          TunePitch();

        }

        g_modeButton.Update();

        if (g_modeButton.JustPressed())
        {
          // Exit current mode
          switch (m_mode)
          {
            case TunerMode::Midi: NoteOff(); break;
          }

          // Cycle modes
          m_mode = static_cast<TunerMode::Type>((m_mode + 1) % TunerMode::Max);

          // Enter new mode

          switch (m_mode)
          {
            case TunerMode::Midi:
              {
                m_newMidiNote = false;
                m_lastMaxAmplitude = 0;
              }
              break;
          }
        }



        float instantFrequency = DetermineSignalPitch();

        m_tunerNote = GetMidiNoteIndexForFrequency(instantFrequency);
        float centerFrequency = GetFrequencyForMidiNoteIndex(m_tunerNote);
        float minFrequency = centerFrequency * QUARTERTONE_DOWN;
        float maxFrequency = centerFrequency * QUARTERTONE_UP;

        if (instantFrequency >= 0.0f)
        {
          if ((filteredFrequency >= 0.0f) && (filteredFrequency >= minFrequency) && (filteredFrequency <= maxFrequency))
          {
            static float const RATE = 0.1f;
            filteredFrequency = (1.0f - RATE) * filteredFrequency + (RATE * instantFrequency);
          }
          else
          {
            filteredFrequency = instantFrequency;
          }
        }
        else
        {
          filteredFrequency = -1.0f;
        }

        // Mini state machine: (disallows voice though - dammit)
        // Detect note strikes
        // only trigger "new note" messages on strikes, or on 3-4 repeatcount
        // only allow semitone changes outside a strike; otherwise display nothing (i.e. false-detect an 330 as a 110, etc.)
        if (m_mode == TunerMode::Midi)
        {
          if (m_maxAmplitude - m_lastMaxAmplitude > 25)
          {
            m_newMidiNote = true;
          }
          if (m_tunerNote >= 0)
          {
            if (m_tunerNote != m_midiNote)
            {
              NoteOff();
              m_midiNote = m_tunerNote;
              noteRepeatCount = 0;
            }
            else
            {
              ++noteRepeatCount;
            }

            if ((m_newMidiNote && (noteRepeatCount == 1)) || (!m_newMidiNote & (noteRepeatCount == 3)))
            {
              NoteOn();
              if (m_newMidiNote)
              {
                // Skip the repeat
                noteRepeatCount = 5;
                m_newMidiNote = false;
              }
            }
          }
          else
          {
            NoteOff();
          }
          m_lastMaxAmplitude = m_maxAmplitude;
        }

        float percent = 0.0f;
        if (filteredFrequency < centerFrequency)
        {
          percent = 0.5f * (filteredFrequency - minFrequency) / (centerFrequency - minFrequency);
        }
        else
        {
          percent = 0.5f + 0.5f * (filteredFrequency - centerFrequency) / (maxFrequency - centerFrequency);
        }

        RenderWideGlyphForNote(m_tunerNote);

        switch (m_mode)
        {
          case TunerMode::Midi:
            {
              Serial.println("MIDI-Mode");
              showmsgXY(320, 197, 1, NULL, "MIDI-Out", GREEN);
            }
            break;
          case TunerMode::Tuner:
            {
              showmsgXY(320, 197, 1, NULL, "MIDI-Out", BLACK);
              PrintCenterTick();
              if (filteredFrequency > 0.0f)
              {
                TFT_Tick(percent);


              }
              else
              {
                TFT_Tick(-1);


              }
            }
            break;
        }

      }

      Stop();
      read_button2();
DEBUG("leave while");
    }//Ende void Go()
  private:
    int m_audioPin;
    char m_buffer[BUFFER_SIZE];

    int m_lastTFT_TickX;
    float m_a440;
    int m_midiNote;
    int m_tunerNote;
    int m_maxAmplitude;
    int m_lastMaxAmplitude;
    TunerMode::Type m_mode;
    bool m_newMidiNote;
};
//Tuner Code Ende
//----------------------------------------------------------------------------

void setup(void)
{
  //---------------------------------------------------
  //Setup für Tuner
  pinMode(kStatusPin, OUTPUT);            // sets the digital pin as output
  pinMode(kStatusPin2, OUTPUT);            // sets the digital pin as output
  digitalWrite(kStatusPin, LOW);
  digitalWrite(kStatusPin2, LOW);
  //---------------------------------------------------
  //Setup für Kurbel
  R.begin();
  R.setPeriod(500);
  pinMode(pin, INPUT_PULLUP);
  //---------------------------------------------------
  //Setup für Münzeinwurf
  pinMode(coin, INPUT);


  //---------------------------------------------------
  //Setup für BatterieStatus
  //pinMode(BatPin, INPUT_PULLUP);
  //pinMode(BatPin, INPUT);
  //PulseStatus = false;
  timer1.start();                                     //Initialisierung timer1 - getBattery()

  //---------------------------------------------------
  Serial1.begin(31250);// initialise MIDI output stream
  while (!Serial1);

  Serial2.begin(31250);// initialisiere MIDI Input über MIDI-IN Buchse
  while (!Serial2);

  Serial.begin(57600);
  uint16_t ID = tft.readID();
  while (!Serial) {
    SysCall::yield();
  }
  if (!sd.begin(SD_CONFIG))
  {
    tft.begin(ID);
    tft.setRotation(1);//Landscape
    tft.fillScreen(RED);



    showmsgXY1(100, 140, 3, NULL, "Keine SD-Karte", WHITE, RED);
    delay(1000);

    ext_midi();//Aufruf Unterprogramm MIDI-IN über MIDI-IN Buchse in Endlosschleife bis Software-Reset über Button "RESET"
    //z_tuner();//Aufruf Unterprogramm TUNER in Endlosschleife bis Software-Reset über Button "Zurueck"
    sd.initErrorHalt();//Programm stoppt hier!
  }

  file = sd.open("/");

/*
  //uint16_t ID = tft.readID();
  Serial.print("TFT ID = 0x");
  Serial.println(ID, HEX);
  Serial.println(SD_FAT_TYPE, DEC);
*/


  pinMode(39, OUTPUT); //Relais zum Einschalten der 24V für die Magnete
  digitalWrite(39, LOW);
  tft.begin(ID);
  tft.setRotation(1);            //Landscape
  tft.fillScreen(BLACK);
  tune_btn.initButton(&tft,  410, 100, 80, 40, WHITE, CYAN, BLACK, "TUNE", 2);
  midi_btn.initButton(&tft, 410, 145, 80, 40, WHITE, CYAN, BLACK, "MIDI", 2);
  play_btn.initButton(&tft, 240, 300, 80, 40, WHITE, CYAN, BLACK, "PLAY", 2);
  stop_btn.initButton(&tft, 155, 300, 80, 40, WHITE, CYAN, BLACK, "STOP", 2);
  back_btn.initButton(&tft, 70, 300, 80, 40, WHITE, CYAN, BLACK, "BACK", 2);
  vwd_btn.initButton(&tft, 410, 300, 80, 40, WHITE, CYAN, BLACK, "VWD", 2);
  pause_btn.initButton(&tft, 325, 300, 80, 40, WHITE, CYAN, BLACK, "PAUSE", 2);
  up_btn.initButton(&tft, 410, 205, 80, 40, WHITE, CYAN, BLACK, "UP", 2);
  down_btn.initButton(&tft, 410, 250, 80, 40, WHITE, CYAN, BLACK, "DOWN", 2);

  tune_btn.drawButton(true);
  midi_btn.drawButton(true);
  up_btn.drawButton(true);
  down_btn.drawButton(true);
  play_btn.drawButton(true);
  stop_btn.drawButton(true);
  back_btn.drawButton(true);
  vwd_btn.drawButton(true);
  pause_btn.drawButton(true);

  tft.fillRect(0, 0, 480, 40, BLUE); //Titelleiste Hintergrund

  showmsgXY(15, 28, 1, &FreeSans12pt7b, "        *             Zither Music Box             *", WHITE);
  //showmsgXY(15, 28, 1, &FreeSans12pt7b, "Akkordzither MIDI-PLAYER V1_0_2     2021", WHITE); //Titelleiste Text
  tft.drawRect(10, 80, 355, 195, WHITE); //Playlist Rahmen
  tft.drawFastHLine(10, 190, 355, WHITE);
  showmsgXY(20, 205, 2, NULL, " Aktuell gewaehlter Titel:", WHITE);
  showmsgXY(15, 47, 1, NULL, " Status      Takt       Tempo     Adjust       Batterie      Cycletime  Beat", WHITE);
  showmsgXY(20, 95, 2, NULL, " Lied aus Liste auswaehlen", WHITE);

  plCount = createPlaylistFile();//erstelle Playlist auf SD-Karte
  if (plCount == 0)
  {
    showmsgXY(40, 100, 3, NULL, "Keine MIDI-Files", RED);
    showmsgXY(75, 130, 3, NULL, "auf SD-Karte", RED);
  }

  // initialize MIDIFile
  SMF.begin(&sd);
  SMF.setMidiHandler(midiCallback);
  SMF.setSysexHandler(sysexCallback);
  SMF.setMetaHandler(metaCallback);
  midiSilence();
}
void(* resetFunc) (void) = 0; //Vorbereitung für SoftwareReset (Aufruf mit resetFunc();)
void loop(void)
{

  // only need to look after 2 things - the user interface (TFT_FSM)
  // and the MIDI playing (MIDI_FSM). While playing we have a different
  // mode from choosing the file, so the FSM will run alternately, depending
  // on which state we are currently in.

  timer1.update();                                    //Intervall timer1 ermöglichen - getBattery()


  static seq_state s = TFTSeq;
  static uint32_t lastUpdate = 0;
  static uint32_t lastMillis = 0;
  uint32_t currMillis = millis();

  cycleTime = currMillis - lastMillis;
  if (cycleTime >= 50 || currMillis - lastUpdate > 1000)
  {
    lastUpdate = currMillis;

    char sBuf[10];
    sprintf(sBuf, "C%3d", cycleTime);

    showmsgXY(380, 58, 2, NULL, sBuf, RED);//Ausgabe CycleTime kann später gelöscht werden
  }
  lastMillis = currMillis;

  switch (s)
  {
    //read_button();
    case TFTSeq:  s = tftFSM(s);  break;
    case MIDISeq: s = midiFSM(s); break;
    default: s = TFTSeq;
  }


}

#else  // SPI_DRIVER_SELECT
#error SPI_DRIVER_SELECT must be two in SdFat/SdFatConfig.h
#endif  //SPI_DRIVER_SELECT
